import datetime

import zope.schema.interfaces
import zope.schema.vocabulary
import zope.publisher.browser
from zope.app.pagetemplate import viewpagetemplatefile
import z3c.form.interfaces
from z3c.form import field
from z3c.form import form
import z3c.form.browser.checkbox

from collective.singing import interfaces
from collective.singing import subscribe
from collective.singing import MessageFactory as _
from collective.singing import message
from collective.singing.browser import utils
from collective.singing.browser import wizard

class Terms(zope.schema.vocabulary.SimpleVocabulary):
    zope.interface.implements(z3c.form.interfaces.ITerms)

    def getValue(self, token):
        return self.getTermByToken(token).value

class ChooseFormatStep(wizard.Step):
    prefix = 'format'

    @property
    def fields(self):
        composers = self.context.composers
        terms = Terms([
            zope.schema.vocabulary.SimpleTerm(
                name, title=composers[name].title)
            for name in sorted(composers.keys())
            ])

        format = zope.schema.Choice(
            __name__='format',
            title=_(u'Format'),
            vocabulary=terms)

        return field.Fields(format)

class SubscribeStep(wizard.Step):
    prefix = 'composer'

    @property
    def fields(self):
        format = self.parent.format()
        return field.Fields(self.context.composers[format].schema)

    def update(self):
        self.subforms = (CollectorDataForm(self.context, self.request),)
        self.subforms[0].update()
        super(SubscribeStep, self).update()

    def extractData(self):
        sub_data, sub_errors = utils.extract_data_prefixed(self.subforms)
        data, errors = self.widgets.extract()
        data.update(sub_data)
        errors = errors + sub_errors
        return data, errors

class CollectorDataForm(utils.OverridableTemplate, form.Form):
    """A subform for the collector specific data.
    """
    index = viewpagetemplatefile.ViewPageTemplateFile('sub-edit.pt')
    prefix = 'collector'
    label = _(u"Filters")
    ignoreContext = True
    
    @property
    def fields(self):
        collector = self.context.collector
        if collector is not None:
            return field.Fields(collector.schema)
        else:
            return field.Fields()

class Subscribe(wizard.Wizard):
    """The add subscription wizard.

    ``context`` is required to be of type ``IChannel``.
    """
    steps = ChooseFormatStep, SubscribeStep
    label = _(u"Fill in the information below to subscribe.")
    success_message = _(
        u"Thanks for your subscription; "
        u"we sent you a message for confirmation.")
    factory = subscribe.SimpleSubscription

    def format(self):
        return self.before_steps[0].widgets.extract()[0]['format']

    def update_steps(self):
        super(Subscribe, self).update_steps()

        if len(self.before_steps) == 0:
            # If there's only one format, we'll skip the first step
            formats = self.context.composers.keys()
            if len(formats) == 1:
                self.current_index += 1
                format_key = '%s.widgets.%s' % (self.current_step.prefix,
                                                'format')
                self.request.form[format_key] = [formats[0]]
                super(Subscribe, self).update_steps()

    def finish(self, data):
        comp_data = {}
        coll_data = {}
        for key, value in data.items():
            name = key.split('.')[-1]
            if key.startswith('composer.collector.'):
                coll_data[name] = value
            elif key.startswith('composer.'):
                comp_data[name] = value

        # Create the subscription and add it to the channel's
        # subscriptions
        subscription = self.create(comp_data, coll_data)
        self.context.subscriptions[subscription.secret].append(subscription)

        # Ask the composer to render a confirmation message
        composer = self.context.composers[self.format()]
        msg = composer.render_confirmation(subscription)
        status, status_msg = message.dispatch(msg)
        if status != u'sent':
            self.status = _(u"We're sorry, but there seems to be an error. "
                            u"Please try again later.\n"
                            u"(${error})",
                            mapping=dict(error=status_msg))

    def create(self, comp_data, coll_data, factory=None):
        """Create a subscription corresponding to data.

        A fake channel to play with:

          >>> class Channel(object):
          ...     composers = {'html': object()}

          >>> channel = Channel()

        We'll patch our ``_secret`` method for now to return
        something, to avoid the complexity of having to register a
        secret adapter:

          >>> form = Subscribe(channel, None)
          >>> form._secret = lambda data, request: data['e-mail']

        We'll also need to provide data adapters for the simple
        subscription implementation to work:

          >>> from zope import component
          >>> from collective.singing import subscribe
          >>> component.provideAdapter(subscribe.SimpleComposerData)
          >>> component.provideAdapter(subscribe.SimpleCollectorData)
          >>> component.provideAdapter(subscribe.SimpleMetadata)

        We're almost there.  The last missing piece is the ``format``
        attribute of our form.  This is normally retrieved by looking
        at the request.  We'll set the format manually:

          >>> form.format = lambda: 'html'

          >>> s = form.create({'e-mail': 'eggs'}, {'filter': 'underground'})
          >>> s.channel is channel
          True
          >>> s.secret
          'eggs'

        Let's get the composer and collector data back:

          >>> from collective.singing import interfaces
          >>> dict(interfaces.IComposerData(s))
          {'e-mail': 'eggs'}
          >>> dict(interfaces.ICollectorData(s))
          {'filter': 'underground'}

        In addition to these, also metadata was saved:

          >>> m = interfaces.ISubscriptionMetadata(s)
          >>> m['format']
          'html'
          >>> m['date'] # doctest: +ELLIPSIS
          datetime.datetime(...)
          >>> m['pending']
          True
        """
        secret = self._secret(comp_data, self.request)

        metadata = dict(format=self.format(),
                        date=datetime.datetime.now(),
                        pending=True)

        if factory is None:
            factory = self.factory
        return factory(self.context, secret, comp_data, coll_data, metadata)

    def _secret(self, data, request):
        """Convenience method for looking up secrets.
        """
        composer = self.context.composers[self.format()]
        return subscribe.secret(self.context, composer, data, self.request)

class Unsubscribe(utils.OverridableTemplate,
                  zope.publisher.browser.BrowserView):
    def index(self):
        return u"<p>You have been unsubscribed.</p>"

    def __call__(self):
        secret = self.request.form['secret']
        del self.context.subscriptions[secret]
        return self.template()



import datetime
import traceback

import persistent.dict
from BTrees.Length import Length
from ZODB.POSException import ConflictError
from zope import component
from zope import interface
import zope.interface.common.mapping
import zope.lifecycleevent

import zc.queue.interfaces

from collective.singing import interfaces

def dispatch(message):
    dispatcher = interfaces.IDispatch(message.payload)
    try:
        value = dispatcher()
        assert value is not None and len(value) == 2, (
            "Invalid return value from %r: %r" %
            (dispatcher, value))
        status, msg = value
    except ConflictError:
        raise
    except Exception, e:
        # TODO: log
        status = u'error'
        msg = traceback.format_exc(e)

    message.status_message = msg
    message.status = status
    return message.status, message.status_message

class Message(object):
    interface.implements(interfaces.IMessage)

    _status = None
    
    def __init__(self, payload, subscription,
                 status=u'new', status_message=None):
        self.payload = payload
        self.subscription = subscription
        self.status_message = status_message
        self.status = status

    @apply
    def status():
        def get(self):
            return self._status
        def set(self, value):
            assert value in interfaces.MESSAGE_STATES, value
            old_status = self.status
            self._status = value
            zope.event.notify(MessageChanged(self, old_status))
            self.status_changed = datetime.datetime.now()

        return property(get, set)

class MessageQueues(persistent.dict.PersistentDict):
    interface.implements(interfaces.IMessageQueues)

    def __init__(self, *args, **kwargs):
        super(MessageQueues, self).__init__(*args, **kwargs)
        for status in interfaces.MESSAGE_STATES:
            self[status] = zc.queue.Queue()
        self._messages_sent = Length()

    @property
    def messages_sent(self):
        return self._messages_sent()

    def dispatch(self):
        sent = 0
        failed = 0
        
        for name in 'new', 'retry':
            queue = self[name]
            while True:
                try:
                    message = queue.pull()
                    # XXX: I think we shold commit immediately after we
                    # pull from the queue.
                except IndexError:
                    break
                else:
                    status, message = dispatch(message)
                    if status == 'sent':
                        sent += 1
                    else:
                        failed += 1

        self._messages_sent.change(sent)
        return sent, failed

    def flush(self):
        for name in 'error', 'sent':
            queue = self[name]
            try:
                while True:
                    queue.pull()
            except IndexError:
                pass

class MessageChanged(zope.lifecycleevent.ObjectModifiedEvent):
    interface.implements(interfaces.IMessageChanged)

    def __init__(self, object, old_status, *descriptions):
        super(MessageChanged, self).__init__(object, *descriptions)
        self.old_status = old_status

@component.adapter(interfaces.IMessage, interfaces.IMessageChanged)
def queue_message(message, event):
    # We expect the message to be in *no queue* when we receive this
    # event!
    queue = message.subscription.channel.queue
    queue[message.status].put(message)



import datetime

import persistent
from zope import interface
from zope import component

from collective.singing import interfaces
from collective.singing import MessageFactory as _

class UnicodeFormatter(object):
    interface.implements(interfaces.IFormatItem)

    def __init__(self, item):
        self.item = item

    def __call__(self):
        return unicode(self.item)

def getIFormatAdapter(obj, format):
    format = component.queryAdapter(
        obj, interfaces.IFormatItem, name=format)
    if format is None:
        return component.getAdapter(obj, interfaces.IFormatItem)
    else:
        return format

def assemble_messages(channel, items=(), use_collector=True):
    collector = channel.collector
    composers = channel.composers

    queued_messages = 0

    for secret, subscriptions in channel.subscriptions.items():
        for sub in subscriptions:
            subscription_metadata = interfaces.ISubscriptionMetadata(sub)

            if subscription_metadata.get('pending'):
                continue

            # Collect items for subscription
            if collector is not None and use_collector:
                collector_items, cue = collector.get_items(
                    subscription_metadata.get('cue'), sub)
                subscription_metadata['cue'] = cue

                # If there was a collector but no items, we'll skip
                # this subscription:
                if not items and len(collector_items) == 0:
                    continue
            else:
                # If there's no collector, we go on with an empty
                # items list:
                collector_items = ()

            final_items = tuple(items) + tuple(collector_items)

            # First format all items...
            format = subscription_metadata['format']
            final_items = [getIFormatAdapter(item, format)() for item in final_items]
            transforms = component.getAllUtilitiesRegisteredFor(
                interfaces.ITransform)

            # ... then transform them ...
            transformed_items = []
            for item in final_items:
                for transform in transforms:
                    item = transform(item, sub)
                transformed_items.append(item)

            # ... and finally render using the right composer. Note
            # that the message is already queued when we render it.
            composer = composers[format]

            message = composer.render(sub, transformed_items)
            queued_messages +=1

    return queued_messages

class AbstractPeriodicScheduler(object):
    interface.implements(interfaces.IScheduler)

    triggered_last = datetime.datetime(1970, 1, 1)
    active = False
    delta = None

    def tick(self, channel):
        now = datetime.datetime.now()
        if self.active and (now - self.triggered_last >= self.delta):
            return self.trigger(channel)

    def trigger(self, channel):
        self.triggered_last = datetime.datetime.now()
        return assemble_messages(channel)

    def __eq__(self, other):
        return (isinstance(other, AbstractPeriodicScheduler) and
                other.delta == self.delta)

    def __ne__(self, other):
        return not self == other

class DailyScheduler(persistent.Persistent, AbstractPeriodicScheduler):
    title = _(u"Daily scheduler")
    delta = datetime.timedelta(days=1)

class WeeklyScheduler(persistent.Persistent, AbstractPeriodicScheduler):
    title = _(u"Weekly scheduler")
    delta = datetime.timedelta(weeks=1)

schedulers = (WeeklyScheduler, DailyScheduler)



import unittest
from zope.testing import doctest
from zope.component import testing
from zope import component
from zope import interface

def create_eventlog(event=interface.Interface):
    value = []
    @component.adapter(event)
    def log(event):
        value.append(event)
    component.provideHandler(log)
    return value

def test_suite():
    return unittest.TestSuite([

        doctest.DocFileSuite('README.txt'),

        doctest.DocFileSuite(
           'message.txt',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        doctest.DocFileSuite(
            'scheduler.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
        ),

        doctest.DocTestSuite(
           'collective.singing.mail',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        doctest.DocTestSuite(
           'collective.singing.mail',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        ])



from zope import interface
from zope import schema
import zope.component.interfaces
from zope.interface.interfaces import IInterface
from zope.interface.common.mapping import IMapping
from zope.annotation.interfaces import IAnnotatable
import z3c.form.interfaces

class IRequestBasedSecret(interface.Interface):
    """A utility that provides a secret based on the request.
    """
    def __call__(request):
        """Return an ASCII secret.
        """


class IComposerBasedSecret(interface.Interface):
    """An adapter for composer that provides a secret based on the
    user's data used for the composer (see also 'IComposerData').
    """
    
    def secret(data):
        """Return an ASCII secret.

        'data' is a dict that corresponds to the IComposer's schema.
        """


class ISubscription(IAnnotatable):
    """A subscription to a channel.
    """
    
    channel = schema.Object(
        title=u"The channel that we're subscribed to",
        schema=IInterface, # should be really IChannel
        )

    secret = schema.ASCIILine(
        title=u"The subscriber's secret, ideally unique across channels",
        description=u"""\
        Might be a hash of the subscriber's e-mail and a secret on the
        server in case we're dealing with an anonymous subscription.
        Used for unsubscription or for providing an overview of all of
        a subscriber's subscriptions.
        """,
        )

MESSAGE_STATES = [u'new',   # just added
                  u'sent',  # sent successfully
                  u'error', # error while sending
                  u'retry', # error while sending, but retrying
                  ]

class IMessage(interface.Interface):
    """Messages are objects ready for sending.
    """
    payload = schema.Field(
        title=u"The message's payload, e.g. the e-mail message.",
        )

    subscription = schema.Object(
        title=u"Subscription, referenced for bookkeeping purposes only.",
        schema=ISubscription,
        )

    status = schema.Choice(
        title=u"State",
        description=u"IMessageChanged is fired automatically when this is set",
        values=MESSAGE_STATES)

    status_message = schema.Text(
        title=u"Status details",
        required=False)

    status_changed = schema.Datetime(
        title=u"Last time this message changed its status",
        )

class IMessageChanged(zope.lifecycleevent.interfaces.IObjectModifiedEvent):
    """An object event on the message that signals that the status has
    changed.
    """
    old_status = schema.TextLine(title=u"Old status of message")


class IComposerData(IMapping):
    """An adapter from the subscription to data used by the composer.

    Entries correspond to the IComposer's schema.
    """


class ICollectorData(IMapping):
    """An adapter from the subscription to data used by the collector.

    Entries correspond to the ICollector's schema.    
    """


class ISubscriptionMetadata(IMapping):
    """An adapter from the subscription to metadata.

    Entries ``format`` and ``date`` are guaranteed to exist.
    ``format`` is the format that's inherent to the composer, and
    ``date`` is the subscription date of type ``datetime.datetime``.

    May be used for application-specific data.

    XXX: I don't like this.  :)
    """


class IComposer(interface.Interface):
    """Composers will typically provide a user interface that lets you
    modify the look of the message rendered through it.
    """
    
    name = schema.TextLine(
        title=u"The Composer's format, e.g. 'html'",
        )
    

    title = schema.TextLine(
        title=u"The Composer's title, e.g. 'HTML E-Mail'",
        )
    
    schema = schema.Object(
        title=u"A schema instance for use in the subscription form",
        description=u"Values are stored via the IComposerData adapter per "
        "subscriber.",
        schema=IInterface,
        )

    def render(subscription, items=()):
        """Given a subscription and a list of items, I will create an
        IMessage and return it.
        """

    def render_confirmation(subscription):
        """Given a subscription, I will create an IMessage that the
        user has to react to in order to confirm the subscription, and
        return it.
        """


class ICollector(interface.Interface):
    """Collectors are useful for automatic newsletters.  They are
    responsible for assembling a list of items for publishing.
    """

    title = schema.TextLine(
        title=u"Title",
        )

    optional = schema.Bool(
        title=u"Subscriber optional",
        )

    schema = schema.Object(
        title=u"A schema instance for use in the subscription form",
        description=u"Values are stored via the ICollectorData adapter per "
        "subscriber.",
        required=False,
        schema=IInterface,
        )

    def get_items(cue=None, subscription=None):
        """Return a tuple '(items, cue)' where 'items' is the items
        that'll be fed into the composer, and 'cue' is an object that
        can be passed to this method again.

        If 'cue' is passed, I will only return items that are newer
        than the items that were returned when the cue was retrieved.

        If 'subscription' is given, I will filter according to the
        subscriber's ICollectorData.
        """

class ITemplate(interface.Interface):
    """Templates are useful for generating and presenting newsletters.
    """

    title = schema.TextLine(
        title=u"Title",
        )

class IScheduler(interface.Interface):
    """A scheduler triggers the sending of messages periodically.
    """
    title = schema.TextLine(
        title=u"Title",
        )

    triggered_last = schema.Datetime(
        title=u"Triggered the last time",
        )

    active = schema.Bool(
        title=u"Active",
        )

    def tick(channel):
        """Check if messages need to be assembled and sent; return the
        number of messages queued or None.
        
        This method is guaranteed to be called periodically.
        """

    def trigger(channel):
        """Assemble and queue messages; return the number of messages
        queued.

        A manual override.
        """

class ISubscriptions(interface.Interface):
    """The `subscriptions` attribute of IChannel implements this.

    Maps secrets to lists of ISubscription objects.
    """
    def __getitem__(key):
        """Return a list of subscriptions for a given key/secret.
        Note that this never raises KeyError.  Instead, if a key is
        not known to the storage, it will return a list.
        """

    def __contains__(key):
        """True if a key/secret is known to the storage.
        """

    def items():
        """Returns an iterator of tuples of the form
        ``(secret, subscriptions)``, where ``subscriptions`` is a
        list of ISubscription objects.
        """

class IMessageQueues(IMapping):
    """A dict that contains one ``zc.queue.interfaces.IQueue`` per
    message status.
    """
    messages_sent = schema.Int(
        title=u"Total number of messages sent through this queue",
        )


class IChannel(interface.Interface):
    """A Channel is what we can subscribe to.

    A Channel is a hub of configuration.  It doesn't do anything by
    itself.  Rather, it provides a number of components to configure
    and work with.  It is also the container of subscriptions to it.
    """

    name = schema.ASCIILine(
        title=u"Unique identifier for this channel across the site.",
        )
    
    title = schema.TextLine(
        title=u"Title",
        )
    
    scheduler = schema.Object(
        title=u"Scheduler (when)",
        required=False,
        schema=IScheduler,
        )

    collector = schema.Object(
        title=u"Collector (what)",
        required=False,
        schema=ICollector,
        )

    composers = schema.Dict(
        title=u"The channel's composers, keyed by format.",
        key_type=schema.TextLine(title=u"Format"),
        value_type=schema.Object(title=u"Composer", schema=IComposer),
        )

    subscriptions = schema.Object(
        title=u"The channel's subscriptions",
        schema=ISubscriptions,
        )

    queue = schema.Object(
        title=u"This channel's message queues, keyed by message status",
        schema=IMessageQueues,
        )


class IFormatItem(interface.Interface):
    """Given an item as retrieved from a collector, this interface
    defines a method that returns a representation of given item ready
    for inclusion in the message via the ``IComposer.render`` method.
    This representation must be of type unicode.
    """

class ITransform(interface.Interface):
    """Allows to rewrite links and the like.
    """
    def __call__(text, subscription):
        """Return transformed text
        """

class IDispatch(interface.Interface):
    """Dispatchers adapt message *payloads* and send them."""

    def __call__():
        """Attempt to send message.

        Must return a tuple ``(status, status_message)``.  See
        ``MESSAGE_STATES`` for possible choices for the status.  The
        status message may be None or a text containing details about
        the status, e.g. why it failed.

        If this method raises an exception, an 'error' is assumed.
        """

class IChannelLookup(interface.Interface):
    """A utility that looks up all channels in a site.
    """

    def __call__():
        """Return a list of ``IChannel`` objects.
        """

class IFormLayer(z3c.form.interfaces.IFormLayer):
    pass



import pprint

import persistent
import persistent.dict
import persistent.list
from zope import interface
from zope import component

from collective.singing import interfaces

def secret(channel, composer, data, request):
    """Look up an appropriate secret.

    Looks for a IRequestBasedSecret first, and falls back to an
    IComposerBasedSecret if none is available or if the
    IRequestBasedSecret component returns None.
    """
    try:
        rbs = component.getUtility(interfaces.IRequestBasedSecret)
    except component.ComponentLookupError, e:
        pass
    else:
        secret = rbs(request)
        if secret is not None:
            return secret
    
    cbs = interfaces.IComposerBasedSecret(composer)
    return cbs.secret(data)

def has_single_format(channel):
    """Returns format of a given channel's single composer, if the
    channel has only one composer; otherwise returns None.
    
      >>> class channel:
      ...     composers = {'sms': object()}
      >>> has_single_format(channel) == 'sms'
      True

      >>> Channel.composers = {}
      >>> has_single_composer(channel) # doctest: +ELLIPSIS
      Traceback (most recent call last):
      ...
      AssertionError...

      >>> Channel.composers = {'sms': None, 'email': None}
      >>> has_single_composer(channel) is None
      True
    """
    composers = channel.composers.keys()
    assert len(composers) >= 1

    if len(composers) == 1:
        return composers[0]

class SimpleSubscriptions(persistent.dict.PersistentDict):
    """A dict that makes sure that when you access an element that it
    doesn't have, you get a nice list back:

      >>> d = SimpleSubscriptions()
      >>> 'bar' in d
      False
      >>> d['bar']
      []
      >>> d['bar'].append('spam')
      >>> d['bar']
      ['spam']
    """
    interface.implements(interfaces.ISubscriptions)

    def __getitem__(self, key):
        if key not in self:
            self[key] = persistent.list.PersistentList()
        return super(SimpleSubscriptions, self).__getitem__(key)

class SimpleSubscription(persistent.Persistent):
    interface.implements(interfaces.ISubscription)
    
    def __init__(self, channel, secret,
                 composer_data, collector_data, metadata):
        self.channel = channel
        self.secret = secret
        self.composer_data.update(composer_data)
        self.collector_data.update(collector_data)
        self.metadata.update(metadata)

    @property
    def composer_data(self):
        return interfaces.IComposerData(self)

    @property
    def collector_data(self):
        return interfaces.ICollectorData(self)

    @property
    def metadata(self):
        return interfaces.ISubscriptionMetadata(self)
        
    def __repr__(self):
        def dict_format(data):
            return pprint.pformat(dict(data)).replace('\n', '')

        data = dict(channel=self.channel)
        for attr in ('composer_data', 'collector_data', 'metadata'):
            data[attr] = dict_format(getattr(self, attr))

        fmt_str = "<SimpleSubscription to %(channel)r with composerdata: %(composer_data)s, collectordata: %(collector_data)s, and metadata: %(metadata)s>"
        return fmt_str % data

def _data_dict(subscription, name):
    data = getattr(subscription, name, None)
    if data is None:
        data = persistent.dict.PersistentDict()
        setattr(subscription, name, data)
    return data

@interface.implementer(interfaces.IComposerData)
@component.adapter(SimpleSubscription)
def SimpleComposerData(subscription):
    return _data_dict(subscription, '_composer_data')

@interface.implementer(interfaces.ICollectorData)
@component.adapter(SimpleSubscription)
def SimpleCollectorData(subscription):
    return _data_dict(subscription, '_collector_data')

@interface.implementer(interfaces.ISubscriptionMetadata)
@component.adapter(SimpleSubscription)
def SimpleMetadata(subscription):
    return _data_dict(subscription, '_metadata')



import zope.publisher.browser
from zope.app.pagetemplate import viewpagetemplatefile

class Macros(zope.publisher.browser.BrowserView):
    template = viewpagetemplatefile.ViewPageTemplateFile('macros.pt')
    
    def __getitem__(self, key):
        return self.template.macros[key]



import email.Message
import email.Parser
import formatter
import htmllib
import mimetools
import MimeWriter
import quopri
import StringIO
import traceback

from zope import interface
from zope import component
import zope.sendmail.interfaces

from collective.singing import interfaces

def header(text, encoding):
    try:
        return text.encode()
    except UnicodeEncodeError:
        encoded_subj = quopri.encodestring(
            text.encode(encoding), header=True)
        return '=?%s?Q?%s?=' % (encoding.upper(), encoded_subj)

def create_html_mail(subject, html, text=None, from_addr=None, to_addr=None,
                     encoding='UTF-8'):
    """Create a mime-message that will render HTML in popular
    MUAs, text in better ones.

    Ripped from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/67083
    """
    html = html.encode(encoding)
    if text is None:
        # Produce an approximate textual rendering of the HTML string,
        # unless you have been given a better version as an argument
        textout = StringIO.StringIO()
        formtext = formatter.AbstractFormatter(formatter.DumbWriter(textout))
        parser = htmllib.HTMLParser(formtext)
        parser.feed(html)
        parser.close()
        text = textout.getvalue()
        del textout, formtext, parser
    else:
        text = text.encode(encoding)
        
    out = StringIO.StringIO() # output buffer for our message
    htmlin = StringIO.StringIO(html) # input buffer for the HTML
    txtin = StringIO.StringIO(text) # input buffer for the plain text
    writer = MimeWriter.MimeWriter(out)

    # Set up some basic headers.  Place subject here because
    # smtplib.sendmail expects it to be in the message, as relevant
    # RFCs prescribe.
    writer.addheader("Subject", header(subject, encoding))
    writer.addheader("MIME-Version", "1.0")
    if from_addr:
        writer.addheader("From", from_addr)
    if to_addr:
        writer.addheader("To", to_addr)

    # Start the multipart section of the message.
    # Multipart/alternative seems to work better on some MUAs than
    # multipart/mixed.
    writer.startmultipartbody("alternative")
    writer.flushheaders()

    # Make plain text section quoted-printable
    subpart = writer.nextpart()
    subpart.addheader("Content-Transfer-Encoding", "quoted-printable")
    pout = subpart.startbody("text/plain", [("charset", encoding)])
    mimetools.encode(txtin, pout, 'quoted-printable')
    txtin.close()

    # The HTML subpart is quoted-printable, too
    subpart = writer.nextpart()
    subpart.addheader("Content-Transfer-Encoding", "quoted-printable")
    pout = subpart.startbody("text/html", [("charset", encoding)])
    mimetools.encode(htmlin, pout, 'quoted-printable')
    htmlin.close()

    # You're done; close your writer and return the message as a string
    writer.lastpart()
    msg = out.getvalue().encode()
    out.close()

    parser = email.Parser.Parser()
    return parser.parsestr(msg)

class Dispatch(object):
    """An IDispatcher registered for ``email.message.Message`` that'll
    send e-mails using ``zope.sendmail``.

    To send a test e-mail, we'll first create an e-mail object:

      >>> message = email.Message.Message()
      >>> message['From'] = 'daniel@testingunderground.com'
      >>> message['To'] = 'plone-users@lists.sourceforge.net'
      >>> message.set_payload('Hello, Plone users!')

    ``Dispatch`` adapts ``email.Message.Message``:
     
      >>> dispatcher = Dispatch(message)
     
    Sending a message without a configured ``IMailDelivery`` will fail:
     
      >>> dispatcher() # doctest: +ELLIPSIS
      Traceback (most recent call last):
      ...
      ComponentLookupError: (<InterfaceClass zope.sendmail.interfaces.IMailDelivery>, '')

    Let's provide our own ``IMailDelivery`` and see what happens:

      >>> class MyMailDelivery(object):
      ...     interface.implements(zope.sendmail.interfaces.IMailDelivery)
      ...
      ...     def send(self, from_, to, message):
      ...         print 'From: ', from_
      ...         print 'To: ', to
      ...         print 'Message follows:'
      ...         print message

      >>> component.provideUtility(MyMailDelivery())
      >>> dispatcher()
      From:  daniel@testingunderground.com
      To:  plone-users@lists.sourceforge.net
      Message follows:
      From: daniel@testingunderground.com
      To: plone-users@lists.sourceforge.net
      <BLANKLINE>
      Hello, Plone users!
      (u'sent', None)

    Note that the last line is the return value.

    If the delivery fails, we'll get a return value with u'error' as
    the first element.

      >>> class MyException(Exception):
      ...     pass
      >>> class MyFailingMailDelivery(object):
      ...     interface.implements(zope.sendmail.interfaces.IMailDelivery)
      ...
      ...     def send(self, from_, to, message):
      ...         raise MyException('This is a test')
      >>> component.provideUtility(MyFailingMailDelivery())
      >>> status, message = dispatcher()
      >>> status
      u'error'
      >>> print message # doctest: +NORMALIZE_WHITESPACE
      Traceback (most recent call last):
      MyException: This is a test
    """
    
    interface.implements(interfaces.IDispatch)
    component.adapts(email.Message.Message)

    def __init__(self, message):
        self.message = message

    def __call__(self):
        msg = self.message
        delivery = component.getUtility(zope.sendmail.interfaces.IMailDelivery)
        try:
            delivery.send(msg['From'], msg['To'], msg.as_string())
        except Exception, e:
            # TODO: log
            return u'error', traceback.format_exc(e)
        else:
            return u'sent', None



from setuptools import setup, find_packages

version = '0.1'

setup(name='collective.singing',
      version=version,
      description="A Zope library for sending notifications and newsletters",
      long_description="""\
""",

      classifiers=[
        "Framework :: Plone",
        "Framework :: Zope2",
        "Framework :: Zope3",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='zope plone notification newsletter',
      author='Daniel Nouri, Thomas Clement Mogensen and contributors',
      author_email='daniel.nouri@gmail.com',
      url='http://pypi.python.org/pypi/collective.singing',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['collective'],
      include_package_data=True,
      zip_safe=False,

      # If the dependency to z3c.form gives you trouble within a Zope
      # 2 environment, try the `fakezope2eggs` recipe
      install_requires=[
          'setuptools',
          'z3c.form',
          'zc.queue',
      ],
      dependency_links=[
      ],
      
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



import unittest
from zope.testing import doctest

from zope import component
import zope.traversing.adapters
import zope.traversing.namespace
from zope.component import testing
import zope.publisher.interfaces.browser
import z3c.form.testing

from collective.singing import subscribe
from collective.singing.browser import macros

def setup_defaults():
    # Set up z3c.form defaults
    z3c.form.testing.setupFormDefaults()
    
    # Make traversal work; register both the default traversable
    # adapter and the ++view++ namespace adapter
    component.provideAdapter(
        zope.traversing.adapters.DefaultTraversable, [None])
    component.provideAdapter(
        zope.traversing.namespace.view, (None, None), name='view')

    # Setup the data adapters for the simple subscription
    component.provideAdapter(subscribe.SimpleComposerData)
    component.provideAdapter(subscribe.SimpleCollectorData)
    component.provideAdapter(subscribe.SimpleMetadata)

    # Setup singing macros
    component.provideAdapter(
        macros.Macros,
        (None, None),
        zope.publisher.interfaces.browser.IBrowserView,
        name='singing-macros')

def test_suite():
    return unittest.TestSuite([

        doctest.DocFileSuite(
           'crud.txt',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        doctest.DocFileSuite(
           'wizard.txt',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        doctest.DocFileSuite(
           'subscribe.txt',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        doctest.DocTestSuite(
           'collective.singing.browser.subscribe',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),

        doctest.DocTestSuite(
           'collective.singing.browser.crud',
           setUp=testing.setUp, tearDown=testing.tearDown,
           ),
        ])
